package net.awired.ajsl.persistence.dao;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.Collection;
import java.util.List;
import javax.persistence.EntityManager;
import javax.persistence.EntityNotFoundException;
import javax.persistence.NoResultException;
import javax.persistence.NonUniqueResultException;
import javax.persistence.PersistenceContext;
import javax.persistence.TypedQuery;
import javax.persistence.criteria.CriteriaBuilder;
import javax.persistence.criteria.CriteriaQuery;
import javax.persistence.criteria.Predicate;
import javax.persistence.criteria.Root;
import net.awired.ajsl.core.lang.reflect.ReflectTool;
import net.awired.ajsl.persistence.entity.Activable;
import net.awired.ajsl.persistence.entity.Archivable;
import net.awired.ajsl.persistence.entity.Defaultable;
import net.awired.ajsl.persistence.entity.IdEntity;
import net.awired.ajsl.persistence.entity.Nested;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public class GenericDAOImpl<ENTITY extends IdEntity<KEY>, KEY extends Serializable> implements
        AbstractDAO<ENTITY, KEY> {

    private static final Logger LOG = LoggerFactory.getLogger(GenericDAOImpl.class);

    @PersistenceContext
    protected EntityManager     entityManager;

    protected Class<ENTITY>     entityClass;
    protected Class<KEY>        keyClass;

    @SuppressWarnings("unchecked")
    public GenericDAOImpl() {
        entityClass = (Class<ENTITY>) ReflectTool.getClassDeclaredInSuperClassGeneric(this, 0);
        keyClass = (Class<KEY>) ReflectTool.getClassDeclaredInSuperClassGeneric(this, 1);
        LOG.debug("Creation of dao for entity '{}' and key type '{}'", entityClass, keyClass);
        checkGenericClass();
    }

    public GenericDAOImpl(Class<ENTITY> entityClass, Class<KEY> keyClass) {
        this.entityClass = entityClass;
        this.keyClass = keyClass;
        LOG.debug("Creation of dao for entity '{}' and key type '{}'", entityClass, keyClass);
        checkGenericClass();
    }

    private String extractBooleanPropertyName(Class<?> clazz) {
        Method[] declaredMethods = clazz.getDeclaredMethods();
        for (Method method : declaredMethods) {
            if (method.getName().startsWith("is")) {
                String res = Character.toLowerCase(method.getName().charAt(2)) + method.getName().substring(3);
                return res;
            }
        }
        throw new RuntimeException("Getter not found in flag Class " + clazz);
    }

    //////////////////
    //////////////////

    @Override
    public ENTITY find(KEY id) throws EntityNotFoundException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Finding by id " + entityClass + "#" + id);
        }
        ENTITY entity = entityManager.find(entityClass, id);
        if (entity == null) {
            throw new EntityNotFoundException("entity " + entityClass + "#" + id + " not found");
        }
        return (ENTITY) entity;
    }

    @Override
    public List<ENTITY> find(Collection<KEY> ids) {
        return find(ids, (Class<?>) null);
    }

    @Override
    public List<ENTITY> find(Collection<KEY> ids, Class<?>... flags) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Finding by ids " + entityClass + "#" + ids + " with flags " + flags);
        }
        CriteriaBuilder cb = entityManager.getCriteriaBuilder();
        CriteriaQuery<ENTITY> criteriaQuery = cb.createQuery(entityClass);
        Root<ENTITY> from = criteriaQuery.from(entityClass);
        criteriaQuery.where(from.in(ids));
        for (Class<?> clazz : flags) {
            if (clazz != null) {
                String propertyName = extractBooleanPropertyName(clazz);

                Predicate nameEquals = cb.equal(from.get(propertyName), true);
                criteriaQuery.where(nameEquals);

                //                Metamodel m = entityManager.getMetamodel();
                //                EntityType<ENTITY> entityType = m.entity(entityClass);
                //                Root<ENTITY> root = criteriaQuery.from(entityClass);
                //                criteriaQuery.where(cb.equal(root.get(propertyName), true));
            }
        }

        TypedQuery<ENTITY> query = entityManager.createQuery(criteriaQuery);
        return query.getResultList();
    }

    @Override
    public List<ENTITY> find(KEY[] ids) {
        return find(Arrays.asList(ids));
    }

    @Override
    public List<ENTITY> findAll() {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Finding all " + entityClass);
        }
        CriteriaBuilder criteriaBuilder = entityManager.getCriteriaBuilder();
        CriteriaQuery<ENTITY> criteriaQuery = criteriaBuilder.createQuery(entityClass);
        criteriaQuery.from(entityClass);
        TypedQuery<ENTITY> query = entityManager.createQuery(criteriaQuery);
        return query.getResultList();
    }

    @Override
    public void persist(ENTITY object) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Persisting " + object);
        }
        entityManager.persist(object);
    }

    @Override
    public ENTITY merge(ENTITY object) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Merging " + object);
        }
        return entityManager.merge(object);
    }

    @Override
    public ENTITY save(ENTITY object) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Saving " + object);
        }
        if (object.getId() != null) {
            return entityManager.merge(object);
        } else {
            entityManager.persist(object);
            return object;
        }
    }

    @Override
    public void save(Collection<ENTITY> objects) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Saving objects " + objects);
        }
        for (ENTITY object : objects) {
            save(object);
        }
    }

    @Override
    public void save(final ENTITY... objects) {
        save(Arrays.asList(objects));
    }

    protected List<ENTITY> findByCriteria(CriteriaQuery<ENTITY> criteriaQuery) {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Finding by criteria " + criteriaQuery);
        }
        return entityManager.createQuery(criteriaQuery).getResultList();
    }

    protected ENTITY findUniqueByCriteria(CriteriaQuery<ENTITY> criteriaQuery) throws NonUniqueResultException,
            NoResultException {
        if (LOG.isDebugEnabled()) {
            LOG.debug("Finding single result by criteria " + criteriaQuery);
        }
        return entityManager.createQuery(criteriaQuery).getSingleResult();
    }

    //    @Override
    //    public List<ENTITY> findByExample(ENTITY example) {
    //        Criteria criteria = Criteria.forClass(entityClass);
    //
    //        Field[] fields = example.getClass().getDeclaredFields();
    //
    //        for (Field field : fields) {
    //            if (field.getName().equals(ENTITY.ID)) {
    //                continue;
    //            }
    //            if (field.getName().equals(IActivable.P_IS_ACTIVE)) {
    //                continue;
    //            }
    //            if (field.getName().equals(IDefaultable.P_IS_DEFAULT)) {
    //                continue;
    //            }
    //            if (field.getName().equals(IHiddenable.P_IS_HIDDEN)) {
    //                continue;
    //            }
    //            if (!field.isAnnotationPresent(Column.class) && !field.isAnnotationPresent(Basic.class)) {
    //                continue;
    //            }
    //
    //            Object value = null;
    //
    //            try {
    //                field.setAccessible(true);
    //                value = field.get(example);
    //            } catch (IllegalArgumentException e) {
    //                continue;
    //            } catch (IllegalAccessException e) {
    //                continue;
    //            }
    //
    //            if (value == null) {
    //                continue;
    //            }
    //
    //            criteria.add(Restrictions.eq(field.getName(), value));
    //        }
    //
    //        if (example instanceof IHiddenable) {
    //            if (((IInheritable) example).getParent() == null) {
    //                criteria.add(Restrictions.isNull(IInheritable.P_PARENT));
    //            } else {
    //                criteria.add(Restrictions.eq(IInheritable.P_PARENT, ((IInheritable) example).getParent()));
    //            }
    //        }
    //
    //        return findByCriteria(criteria);
    //    }

    //    @Override
    //    public void delete(final KEY id) {
    //        delete(find(id));
    //    }
    //
    //    @Override
    //    public void delete(final KEY... ids) {
    //        deleteAll(get(ids), true);
    //    }
    //
    //    @Override
    //    public void delete(final ENTITY object) {
    //        if (isDefaultable) {
    //            checkIfDefault(object);
    //        }
    //        if (isHiddenable) {
    //            ((IHiddenable) object).setHidden(true);
    //            entityManager.merge(object);
    //        } else {
    //            entityManager.remove(object);
    //        }
    //    }
    //
    //    @Override
    //    public void delete(final ENTITY... objects) throws UnsupportedOperationException {
    //        deleteAll(Arrays.asList(objects), true);
    //    }
    //
    //    @Override
    //    public void deleteAll() throws UnsupportedOperationException {
    //        deleteAll(findAll(), false);
    //    }
    //
    //    private void deleteAll(final Collection<ENTITY> objects, boolean checkIdDefault)
    //            throws UnsupportedOperationException {
    //        if (checkIdDefault) {
    //            if (isDefaultable) {
    //                for (ENTITY object : objects) {
    //                    checkIfDefault(object);
    //                }
    //            }
    //        }
    //        if (isHiddenable) {
    //            for (ENTITY object : objects) {
    //                ((IHiddenable) object).setHidden(true);
    //                entityManager.merge(object);
    //            }
    //        } else {
    //            for (ENTITY object : objects) {
    //                entityManager.remove(object);
    //            }
    //        }
    //    }

    @Override
    public void refresh(ENTITY entity) {
        entityManager.refresh(entity);
    }

    //    @Override
    //    public void flushAndClear() {
    //        entityManager.flush();
    //        entityManager.clear();
    //    }

    ///////////////
    //////////////

    //    private void checkIfDefault(ENTITY entity) {
    //        if (((IDefaultable) entity).isDefault()) {
    //            throw new UnsupportedOperationException("can't delete default entity " + entityClass + "#"
    //                    + entity.getId());
    //        }
    //    }
    //
    //    @Override
    //    public boolean isActivable() {
    //        return isActivable;
    //    }
    //
    //    @Override
    //    public boolean isDefaultable() {
    //        return isDefaultable;
    //    }
    //
    //    @Override
    //    public boolean isHiddenable() {
    //        return isHiddenable;
    //    }
    //
    //    @Override
    //    public boolean isInheritable() {
    //        return isInheritable;
    //    }
    //
    //    @Override
    //    public void setAsDefault(IDefaultable object) {
    //        if (object.getExample() != null) {
    //            List<ENTITY> objects = findByExample((ENTITY) object.getExample());
    //            for (ENTITY o : objects) {
    //                if (object != o) {
    //                    ((IDefaultable) o).setDefault(false);
    //                    entityManager.merge(o);
    //                }
    //            }
    //        }
    //        object.setDefault(true);
    //
    //        if (((ENTITY) object).getId() != null) {
    //            entityManager.merge(object);
    //        } else {
    //            entityManager.persist(object);
    //        }
    //    }
    //
    private void checkGenericClass() {
        for (Class i : entityClass.getInterfaces()) {
            if (i == Defaultable.class) {
                isDefaultable = true;
            } else if (i == Activable.class) {
                isActivable = true;
            } else if (i == Archivable.class) {
                isHiddenable = true;
            } else if (i == Nested.class) {
                isInheritable = true;
            }
        }
    }

    //
    //    @Override
    //    public List<ENTITY> find(IInheritable<ENTITY> parent) {
    //        if (parent == null) {
    //            return findByCriteria(Criteria.forClass(entityClass).add(Restrictions.isNull(IInheritable.P_PARENT)));
    //        } else {
    //            return findByCriteria(Criteria.forClass(entityClass).add(Restrictions.eq(IInheritable.P_PARENT, parent)));
    //        }
    //    }
    //
    private boolean isDefaultable;

    private boolean isActivable;

    private boolean isHiddenable;

    private boolean isInheritable;

}
